# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_split
from ....common._registration import register_converter


def convert_split(scope, operator, container):
    # ONNX Split may evenly divide the input along the specified axis if "split" attribute is not specified.
    # Also, CoreML always evenly split the input along C-axis. Consequently, we only need to specify the axis
    # and make sure the number of outputs in ONNX matches that in CoreML.
    apply_split(scope, operator.input_full_names, operator.output_full_names, container,
                operator_name=operator.full_name, axis=1)


register_converter('split', convert_split)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter
from .SVC import extract_support_vectors_as_dense_tensor


def convert_svm_regressor(scope, operator, container):
    params = operator.raw_operator.supportVectorRegressor

    kernel_enum = {'linearKernel': 'LINEAR', 'polyKernel': 'POLY',
                   'rbfKernel': 'RBF', 'sigmoidKernel': 'SIGMOID', 'precomputedKernel': 'PRECOMPUTED'}
    kernel = params.kernel
    kernel_val = kernel.WhichOneof('kernel')
    svr_kernel = kernel_enum[kernel_val]

    if kernel_val == 'rbfKernel':
        svr_kernel_params = [kernel.rbfKernel.gamma, 0.0, 0.0]
    elif kernel_val == 'polyKernel':
        svr_kernel_params = [kernel.polyKernel.gamma,
                             kernel.polyKernel.coef0, kernel.polyKernel.degree]
    elif kernel_val == 'sigmoidKernel':
        svr_kernel_params = [kernel.sigmoidKernel.gamma,
                             kernel.sigmoidKernel.coef0, 0.0]
    elif kernel_val == 'linearKernel':
        svr_kernel_params = [0.0, 0.0, 0.0]

    n_supports, support_vectors = extract_support_vectors_as_dense_tensor(operator.raw_operator.supportVectorRegressor)

    svr_coefficients = params.coefficients.alpha
    if isinstance(params.rho, list):
        svr_rho = [-x for x in params.rho]
    else:
        svr_rho = [-params.rho]

    op_type = 'SVMRegressor'
    op_name = scope.get_unique_operator_name(op_type)
    attrs = {'name': op_name}
    attrs['kernel_type'] = svr_kernel
    attrs['kernel_params'] = svr_kernel_params
    attrs['support_vectors'] = support_vectors
    attrs['n_supports'] = n_supports
    attrs['coefficients'] = svr_coefficients
    attrs['rho'] = svr_rho

    container.add_node(op_type, operator.input_full_names, operator.output_full_names, op_domain='ai.onnx.ml', **attrs)


register_converter('supportVectorRegressor', convert_svm_regressor)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_shape_calculator
from ...common.data_types import FloatTensorType, Int64TensorType, Int64Type, StringTensorType, StringType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculte_tensor_to_label_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C] ---> [N, 1]

    Note that N must be 1 currently because TensorToProbability doesn't support batch size larger than 1.
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    N = operator.inputs[0].type.shape[0]
    if operator.target_opset < 7:
        output_shape = [1, 1]
    else:
        output_shape = [N, 1]

    if type(operator.outputs[0].type) in [Int64Type, Int64TensorType]:
        operator.outputs[0].type = Int64TensorType(output_shape, doc_string=operator.outputs[0].type.doc_string)
    elif type(operator.outputs[0].type) in [StringType, StringTensorType]:
        operator.outputs[0].type = StringTensorType(output_shape, doc_string=operator.outputs[0].type.doc_string)
    else:
        raise ValueError('Unsupported label type')


register_shape_calculator('tensorToLabel', calculte_tensor_to_label_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import copy
from onnx import onnx_pb as onnx_proto
from ...common._apply_operation import apply_concat, apply_cast
from ...common.data_types import Int64TensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._registration import register_converter, register_shape_calculator


def convert_sparkml_polynomial_expansion(scope, operator, container):
    op = operator.raw_operator
    feature_count = operator.inputs[0].type.shape[1]
    degree = op.getDegree()
    all_combinations = calc_combinations(feature_count, degree)
    transformed_columns = []

    for i, comb in enumerate(all_combinations):
        if comb is None:
            pass
        else:
            comb_name = scope.get_unique_variable_name('comb')
            col_name = scope.get_unique_variable_name('col')
            prod_name = scope.get_unique_variable_name('prod')

            container.add_initializer(comb_name, onnx_proto.TensorProto.INT64,
                                      [len(comb)], list(comb))

            container.add_node(
                'ArrayFeatureExtractor',
                [operator.inputs[0].full_name, comb_name], col_name,
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
                op_domain='ai.onnx.ml')
            reduce_prod_input = col_name
            if (operator.inputs[0].type._get_element_onnx_type()
                    == onnx_proto.TensorProto.INT64):
                float_col_name = scope.get_unique_variable_name('col')
                container.add_node('Cast', col_name, float_col_name,
                                   name=scope.get_unique_operator_name('Cast'),
                                   to=onnx_proto.TensorProto.FLOAT)
                reduce_prod_input = float_col_name

            container.add_node(
                'ReduceProd', reduce_prod_input, prod_name,
                axes=[1], name=scope.get_unique_operator_name('ReduceProd'))
            transformed_columns.append(prod_name)

    if (operator.inputs[0].type._get_element_onnx_type()
            == onnx_proto.TensorProto.INT64):
        concat_result_name = scope.get_unique_variable_name('concat_result')

        apply_concat(scope, [t for t in transformed_columns], concat_result_name,
                     container, axis=1)
        apply_cast(scope, concat_result_name, operator.outputs[0].full_name,
                   container,
                   to=onnx_proto.TensorProto.INT64)
    else:
        apply_concat(scope, [t for t in transformed_columns], operator.outputs[0].full_name,
                     container, axis=1)


register_converter('pyspark.ml.feature.PolynomialExpansion', convert_sparkml_polynomial_expansion)


def expand_inner(
      values,
      last_index,
      degree,
      multiplier,
      poly_values,
      cur_poly_index):
    if degree == 0 or last_index < 0:
        if cur_poly_index >= 0:
            # skip the very first 1
            poly_values[cur_poly_index] = multiplier.copy()
    else:
        v = values[last_index]
        last_index1 = last_index - 1
        alpha = multiplier.copy()
        i = 0
        cur_start = cur_poly_index
        while i <= degree:
            cur_start, multiplier = expand_inner(values, last_index1, degree - i, alpha, poly_values, cur_start)
            i += 1
            alpha.append(v)
    return cur_poly_index + get_poly_size(last_index + 1, degree), multiplier


def calc_combinations(feature_count, degree):
    values = range(0, feature_count)
    poly_size = get_poly_size(feature_count, degree)
    poly_values = [None] * poly_size
    expand_inner(values, feature_count - 1, degree, [], poly_values, -1)
    return poly_values


def get_poly_size(feature_count, degree):
    return get_combinations_count(feature_count+degree, degree)


def get_combinations_count(n, k):
    from math import factorial
    if n == k or k == 0:
        return 1
    if k == 1 or k == n - 1:
        return n
    return factorial(n) // factorial(k) // factorial(n - k)


def calculate_sparkml_polynomial_expansion_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[
                                 FloatTensorType, Int64TensorType])

    N = operator.inputs[0].type.shape[0]
    C = get_combinations_count(operator.inputs[0].type.shape[1], operator.raw_operator.getDegree())
    operator.outputs[0].type = copy.deepcopy(operator.inputs[0].type)
    operator.outputs[0].type.shape = [N, C]


register_shape_calculator('pyspark.ml.feature.PolynomialExpansion', calculate_sparkml_polynomial_expansion_output_shapes)



# SPDX-License-Identifier: Apache-2.0

import unittest
import sys
import inspect
import os
import time
import pathlib
import numpy
import pandas
from pyspark.ml import Pipeline, PipelineModel
from pyspark.ml.classification import LogisticRegression
from pyspark.ml.feature import StringIndexer, OneHotEncoder, VectorAssembler
from onnxmltools import convert_sparkml
from onnxmltools.convert.sparkml import buildInitialTypesSimple, buildInputDictSimple
from onnxmltools.utils.utils_backend import OnnxRuntimeAssertionError, compare_outputs
from onnxmltools.utils.utils_backend_onnxruntime import run_with_runtime, _compare_expected
from tests.sparkml import SparkMlTestCase


class ProfileSparkmlPipeline(SparkMlTestCase):
    def _get_spark_options(self):
        # add additional jar files before creating SparkSession
        return {'spark.jars.packages': "ml.combust.mleap:mleap-spark_2.11:0.13.0"}


class ProfileSparkmlPipeline(SparkMlTestCase):

    def test_profile_sparkml_pipeline(self):
        import mleap.pyspark
        from mleap.pyspark.spark_support import SimpleSparkSerializer

        # add additional jar files before creating SparkSession
        this_script_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
        input_path = os.path.join(this_script_dir, "data", "AdultCensusIncomeOriginal.csv")
        full_data = self.spark.read.format('csv') \
            .options(header='true', inferschema='true').load(input_path)
        training_data, test_data = full_data.randomSplit([0.9, 0.1], seed=1)

        label = "income"
        dtypes = dict(training_data.dtypes)
        dtypes.pop(label)

        si_xvars = []
        ohe_xvars = []
        feature_cols = []
        for idx, key in enumerate(dtypes):
            if dtypes[key] == "string":
                feature_col = "-".join([key, "encoded"])
                feature_cols.append(feature_col)

                tmp_col = "-".join([key, "tmp"])
                si_xvars.append(StringIndexer(inputCol=key, outputCol=tmp_col, handleInvalid="skip"))
                ohe_xvars.append(OneHotEncoder(inputCols=[tmp_col], outputCols=[feature_col], dropLast=False))
            else:
                feature_cols.append(key)
        si_label = StringIndexer(inputCol=label, outputCol='label')
        assembler = VectorAssembler(inputCols=feature_cols, outputCol="features")
        lr = LogisticRegression(regParam=0.001)
        pipeline = Pipeline(stages=si_xvars + ohe_xvars + [si_label, assembler, lr])

        # filter out the records which will cause error
        # use only one record for prediction
        test_data = test_data.limit(1)
        # create Spark and Onnx models
        model = pipeline.fit(training_data)
        model_onnx = convert_sparkml(model, 'Sparkml Pipeline', buildInitialTypesSimple(test_data))
        # save Onnx model for runtime usage
        if model_onnx is None: raise AssertionError("Failed to create the onnx model")
        model_path = os.path.join("tests", "profile_pipeline_model.onnx")
        with open(model_path, "wb") as f:
            f.write(model_onnx.SerializeToString())

        # Create MLeap model
        model_zip_path = os.path.join(this_script_dir, "tests", "mleap-pipeline.zip")
        if os.path.exists(model_zip_path):
            os.remove(model_zip_path)
        model_zip_url = "jar:" + pathlib.Path(model_zip_path).as_uri()
        # save the pipeline also in MLeap format
        empty_df = self.spark.createDataFrame([], model.transform(test_data).schema)
        model.serializeToBundle(model_zip_url, empty_df)
        mleap_pipeline = PipelineModel.deserializeFromBundle(model_zip_url)

        spark_times = []
        mleap_times = []
        runtime_times = []
        for i in range(0, 20):
            data_np = buildInputDictSimple(test_data)
            # run the model in Spark
            start = time.time()
            spark_prediction = model.transform(test_data)
            end = time.time()
            spark_times.append(1000 * (end - start))

            # run with MLeap
            start = time.time()
            mleap_prediction = mleap_pipeline.transform(test_data)
            end = time.time()
            mleap_times.append(1000 * (end - start))

            if i == 0:  # compare only once
                _compare_mleap_pyspark(mleap_prediction, spark_prediction)

            # run the model in onnx runtime
            start = time.time()
            output, session = run_with_runtime(data_np, model_path)
            end = time.time()
            runtime_times.append(1000 * (end - start))

            # compare results
            if i == 0:  # compare only once
                expected = [
                    spark_prediction.toPandas().label.values.astype(numpy.float32),
                    spark_prediction.toPandas().prediction.values.astype(numpy.float32),
                    spark_prediction.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(
                        numpy.float32)
                ]
                _compare_expected(expected, output, session, model_path, decimal=5, onnx_shape=None)

        gen_plot(spark_times, mleap_times, runtime_times)


def _compare_mleap_pyspark(mleap_prediction, spark_prediction):
    spark_pandas = spark_prediction.toPandas()
    mleap_pandas = mleap_prediction.toPandas()
    spark_predicted_labels = spark_pandas.prediction.values
    mleap_predicted_labels = mleap_pandas.prediction.values
    msg = compare_outputs(spark_predicted_labels, mleap_predicted_labels, decimal=5)
    if msg:
        raise OnnxRuntimeAssertionError("Predictions in mleap and spark do not match")
    spark_probability = spark_pandas.probability.apply(lambda x: pandas.Series(x.toArray())).values
    mleap_probability = mleap_pandas.probability.apply(lambda x: pandas.Series(x.toArray())).values
    msg = compare_outputs(spark_probability, mleap_probability, decimal=5)
    if msg:
        raise OnnxRuntimeAssertionError("Probabilities in mleap and spark do not match")


def gen_plot(spark_times, mleap_times, runtime_times):
    import matplotlib.pyplot as pyplot
    pyplot.hist(spark_times, label='pyspark')
    pyplot.hist(mleap_times, label='MLeap')
    pyplot.hist(runtime_times, label='onnxruntime')
    pyplot.ylabel('Frequency')
    pyplot.xlabel('Prediction Time(ms)')
    pyplot.legend()
    fig = pyplot.gcf()
    #pyplot.show()
    pyplot.draw()
    fig.savefig('tests/spark-perf-histogram.png')


if __name__ == "__main__":
    unittest.main()




# SPDX-License-Identifier: Apache-2.0

import numpy as np

from ....common._apply_operation import apply_elu, apply_hard_sigmoid, apply_leaky_relu, apply_prelu, apply_relu, \
    apply_sigmoid, apply_tanh, apply_affine, apply_parametric_softplus, apply_scaled_tanh, apply_thresholded_relu
from ....common._registration import register_converter


def convert_activation(scope, operator, container):
    inputs = [variable.full_name for variable in operator.inputs]
    outputs = [variable.full_name for variable in operator.outputs]
    attrs = {'name': operator.full_name}

    params = operator.raw_operator.activation
    activation_type = params.WhichOneof('NonlinearityType')

    # Create ONNX objects by high-level APIs such as apply_relu(...) if possible.
    # Otherwise, we use low-level APIs such as add_node(...)
    if activation_type == 'leakyReLU':
        apply_leaky_relu(scope, inputs, outputs, container, operator_name=attrs['name'], alpha=params.leakyReLU.alpha)
    elif activation_type == 'ReLU':
        apply_relu(scope, inputs, outputs, container, operator_name=attrs['name'])
    elif activation_type == 'PReLU':
        apply_prelu(scope, inputs[0], outputs, container, operator_name=attrs['name'], slope=np.asarray([params.PReLU.alpha.floatValue]))
    elif activation_type == 'ELU':
        apply_elu(scope, inputs, outputs, container, operator_name=attrs['name'], alpha=params.ELU.alpha)
    elif activation_type == 'tanh':
        apply_tanh(scope, inputs, outputs, container, operator_name=attrs['name'])
    elif activation_type == 'sigmoid':
        apply_sigmoid(scope, inputs, outputs, container, operator_name=attrs['name'])
    elif activation_type == 'sigmoidHard':
        apply_hard_sigmoid(scope, inputs, outputs, container, operator_name=attrs['name'],
                           alpha=params.sigmoidHard.alpha, beta=params.sigmoidHard.beta)
    elif activation_type == 'linear':
        apply_affine(scope, inputs[0], outputs[0], container, operator_name=attrs['name'],
                     alpha=params.linear.alpha, beta=params.linear.beta)
    elif activation_type == 'parametricSoftplus':
        apply_parametric_softplus(scope, inputs[0], outputs[0], container, operator_name=attrs['name'],
                                  alpha=params.parametricSoftplus.alpha.floatValue, beta=params.parametricSoftplus.beta.floatValue)
    elif activation_type =='scaledTanh':
        apply_scaled_tanh(scope, inputs[0], outputs[0], container, operator_name=attrs['name'],
                          alpha=params.scaledTanh.alpha, beta=params.scaledTanh.beta)
    elif activation_type == 'thresholdedReLU':
        apply_thresholded_relu(scope, inputs, outputs, container, operator_name=attrs['name'],
                               alpha=params.thresholdedReLU.alpha)
    else:
        if activation_type == 'softsign':
            op_type = 'Softsign'
        elif activation_type == 'softplus':
            op_type = 'Softplus'
        else:
            raise TypeError('Unsupported activation layer {0}'.format(activation_type))

        container.add_node(op_type, inputs, outputs, **attrs)


register_converter('activation', convert_activation)



# SPDX-License-Identifier: Apache-2.0

from onnxconverter_common.data_types import FloatTensorType
from ..common._container import LibSvmModelContainer
from ..common._topology import Topology


def _parse_libsvm_simple_model(scope, model, inputs):
    '''
    This function handles all non-pipeline models.

    :param scope: Scope object
    :param model: A libsvm object (e.g., OneHotEncoder and LogisticRegression)
    :param inputs: A list of variables
    :return: A list of output variables which will be passed to next stage
    '''

    if model.get_svm_type() in (0, 1):
        label_variable = scope.declare_local_variable('label', FloatTensorType())
        probability_map_variable = scope.declare_local_variable('probabilities', FloatTensorType())
        this_operator = scope.declare_local_operator("LibSvmSVC", model)
        this_operator.inputs = inputs
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(probability_map_variable)
    elif model.get_svm_type() in (4, 3):
        # We assume that all scikit-learn operator can only produce a single float tensor.
        variable = scope.declare_local_variable('variable', FloatTensorType())
        this_operator = scope.declare_local_operator("LibSvmSVR", model)
        this_operator.inputs = inputs
        this_operator.outputs.append(variable)
    else:
        raise ValueError("Unknown SVM type '{0}'".format(model.get_svm_type()))
    return this_operator.outputs


def _parse_libsvm(scope, model, inputs):
    '''
    This is a delegate function. It doesn't nothing but invoke the correct parsing function according to the input
    model's type.
    :param scope: Scope object
    :param model: A scikit-learn object (e.g., OneHotEncoder and LogisticRegression)
    :param inputs: A list of variables
    :return: The output variables produced by the input model
    '''
    return _parse_libsvm_simple_model(scope, model, inputs)


def parse_libsvm(model, initial_types=None, target_opset=None,
                 custom_conversion_functions=None,
                 custom_shape_calculators=None):
    # Put svmlib object into an abstract container so that our framework
    # can work seamlessly on models created
    # with different machine learning tools.
    raw_model_container = LibSvmModelContainer(model)

    # Declare a computational graph. It will become a representation of
    # the input scikit-learn model after parsing.
    topology = Topology(raw_model_container, default_batch_size='None',
                        initial_types=initial_types,
                        target_opset=target_opset,
                        custom_conversion_functions=custom_conversion_functions,
                        custom_shape_calculators=custom_shape_calculators)

    # Declare an object to provide variables' and operators' naming mechanism.
    # In contrast to CoreML, one global scope
    # is enough for parsing scikit-learn models.
    scope = topology.declare_scope('__root__')

    # Declare input variables. They should be the inputs of the scikit-learn model
    # you want to convert into ONNX.
    inputs = []
    for var_name, initial_type in initial_types:
        inputs.append(scope.declare_local_variable(var_name, initial_type))

    # The object raw_model_container is a part of the topology we're going to return.
    # We use it to store the inputs of
    # the libsvm's computational graph.
    for variable in inputs:
        raw_model_container.add_input(variable)

    # Parse the input libsvm model as a Topology object.
    outputs = _parse_libsvm(scope, model, inputs)

    # THe object raw_model_container is a part of the topology we're going to return.
    # We use it to store the outputs of
    # the scikit-learn's computational graph.
    for variable in outputs:
        raw_model_container.add_output(variable)

    return topology



# SPDX-License-Identifier: Apache-2.0

from ....common._apply_operation import apply_mul
from ....common._registration import register_converter


def convert_dot(scope, operator, container):
    # To calculate cosine similarity, we first use LpNormalization to make the two input vectors unit-length.
    # Then, we calculate element-wise product of the two unit-length vectors. Finally, the similarity is the
    # sum of all the product's elements. Notice that we carefully specify the axis of the subsequent operators,
    # so they can work properly with a batch of vectors.

    if operator.raw_operator.dot.cosineSimilarity:
        # Normalize the first input and store the result on a temporal variable
        intra_variable_name1 = scope.get_unique_variable_name(operator.inputs[0].full_name + '_normalized')
        normalizer_name1 = scope.get_unique_operator_name('L2NormNormalizer')
        attrs1 = {'name': normalizer_name1, 'p': 2, 'aixs': 1}
        container.add_node('LpNormalization', [operator.inputs[0].full_name], [intra_variable_name1], **attrs1)

        # Normalize the second input and store the result on a temporal variable
        intra_variable_name2 = scope.get_unique_variable_name(operator.inputs[1].full_name + '_normalized')
        normalizer_name2 = scope.get_unique_operator_name('L2NormNormalizer')
        attrs2 = {'name': normalizer_name2, 'p': 2, 'aixs': 1}
        container.add_node('LpNormalization', [operator.inputs[1].full_name], [intra_variable_name2], **attrs2)
    else:
        # This case is a simple dot product; no normalization is required.
        intra_variable_name1 = operator.inputs[0].full_name
        intra_variable_name2 = operator.inputs[1].full_name

    # Do element-wise product of the two unit-length tensors
    product_name = scope.get_unique_variable_name(intra_variable_name1 + '_multiply_' + intra_variable_name2)
    apply_mul(scope, [intra_variable_name1, intra_variable_name2], product_name, container, broadcast=0)

    # Sum up results from different dimensions to get the final cosine similarity
    reducer_name = scope.get_unique_operator_name('ReduceSum')
    reducer_attrs = {'name': reducer_name, 'axes': [1], 'keepdims': 1}
    container.add_node('ReduceSum', [product_name], operator.output_full_names, **reducer_attrs)


register_converter('dot', convert_dot)



# SPDX-License-Identifier: Apache-2.0

import copy
from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_sequence_repeat_output_shapes(operator):
    '''
    Allowed input/output patterns are
        1. [N, C, H, W] ---> [N', C, H, W]
    '''
    check_input_and_output_numbers(operator, input_count_range=1, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input_shape = operator.inputs[0].type.shape
    if len(input_shape) not in [2, 4]:
        raise RuntimeError('Input shape of CoreML SequenceRepeat must be either 2-D or 4-D but got %s' % input_shape)

    output_shape = copy.deepcopy(operator.inputs[0].type.shape)
    if output_shape[0] != None:
        output_shape[0] *= operator.raw_operator.sequenceRepeat.nRepetitions

    operator.outputs[0].type = FloatTensorType(output_shape, doc_string=operator.outputs[0].type.doc_string)


register_shape_calculator('sequenceRepeat', calculate_sequence_repeat_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter


def convert_dictionary_vectorizer(scope, operator, container):
    op_type = 'DictVectorizer'
    attrs = {'name': operator.full_name}
    raw_model = operator.raw_operator.dictVectorizer
    if raw_model.HasField('stringToIndex'):
        attrs['string_vocabulary'] = raw_model.stringToIndex.vector
    else:
        attrs['int64_vocabulary'] = raw_model.int64ToIndex.vector

    container.add_node(op_type, [operator.inputs[0].full_name], [operator.outputs[0].full_name],
                       op_domain='ai.onnx.ml', **attrs)


register_converter('dictVectorizer', convert_dictionary_vectorizer)



# SPDX-License-Identifier: Apache-2.0

from ....common._registration import register_shape_calculator
from ....common.data_types import FloatTensorType
from ....common.utils import check_input_and_output_numbers, check_input_and_output_types


def calculate_bidirectional_lstm_output_shapes(operator):
    '''
    See bidirectional LSTM's conversion function for its output shapes.
    '''
    check_input_and_output_numbers(operator, input_count_range=[1, 5], output_count_range=[1, 5])
    check_input_and_output_types(operator, good_input_types=[FloatTensorType])

    input_shape = operator.inputs[0].type.shape

    # LSTM accepts [N, C] and [N, C, 1, 1] inputs
    if len(input_shape) not in [2, 4]:
        raise RuntimeError('Input must be a 2-D or 4-D tensor')

    params = operator.raw_operator.biDirectionalLSTM
    # The following line is more accurate but it may break some tests
    # output_shape = ['None', params.outputVectorSize] if params.params.sequenceOutput else [1, 2 *params.outputVectorSize]
    output_shape = ['None', 2 * params.outputVectorSize]
    state_shape = [1, params.outputVectorSize]

    # TODO: Changing input shapes of an operator is dangerous, this should be move to Topology's _fix_shapes function
    if len(operator.inputs) > 1:
        Y_h_in = operator.inputs[1]  # The forward initial hidden state of a single sequence
        Y_h_in.type.shape = state_shape
        Y_h_rev_in = operator.inputs[3]  # The backward initial hidden state of a single sequence
        Y_h_rev_in.type.shape = state_shape
    if len(operator.inputs) > 2:
        Y_c_in = operator.inputs[2]  # The forward initial cell state of a single sequence
        Y_c_in.type.shape = state_shape
        Y_c_rev_in = operator.inputs[4]  # The backward initial cell state of a single sequence
        Y_c_rev_in.type.shape = state_shape

    operator.outputs[0].type.shape = output_shape
    if len(operator.outputs) > 1:
        operator.outputs[1].type.shape = state_shape
        operator.outputs[3].type.shape = state_shape
    if len(operator.outputs) > 2:
        operator.outputs[2].type.shape = state_shape
        operator.outputs[4].type.shape = state_shape


register_shape_calculator('biDirectionalLSTM', calculate_bidirectional_lstm_output_shapes)



# SPDX-License-Identifier: Apache-2.0

from uuid import uuid4
import xgboost
import onnx
from ..common.onnx_ex import get_maximum_opset_supported
from ..common._topology import convert_topology
from ._parse import parse_xgboost, WrappedBooster

# Invoke the registration of all our converters and shape calculators
# from . import shape_calculators
from . import operator_converters, shape_calculators


def convert(model, name=None, initial_types=None, doc_string='', target_opset=None,
            targeted_onnx=onnx.__version__, custom_conversion_functions=None,
            custom_shape_calculators=None):
    '''
    This function produces an equivalent ONNX model of the given xgboost model.

    :param model: A xgboost model
    :param initial_types: a python list. Each element is a tuple of a variable name and a type defined in data_types.py
    :param name: The name of the graph (type: GraphProto) in the produced ONNX model (type: ModelProto)
    :param doc_string: A string attached onto the produced ONNX model
    :param target_opset: number, for example, 7 for ONNX 1.2, and 8 for ONNX 1.3.
    :param targeted_onnx: A string (for example, '1.1.2' and '1.2') used to specify the targeted ONNX version of the
        produced model. If ONNXMLTools cannot find a compatible ONNX python package, an error may be thrown.
    :param custom_conversion_functions: a dictionary for specifying the user customized conversion function
    :param custom_shape_calculators: a dictionary for specifying the user customized shape calculator
    :return: An ONNX model (type: ModelProto) which is equivalent to the input xgboost model
    '''
    if initial_types is None:
        raise ValueError('Initial types are required. See usage of convert(...) in \
                           onnxmltools.convert.xgboost.convert for details')
    if name is None:
        name = str(uuid4().hex)

    if isinstance(model, xgboost.Booster):
        model = WrappedBooster(model)
    target_opset = target_opset if target_opset else get_maximum_opset_supported()
    topology = parse_xgboost(model, initial_types, target_opset, custom_conversion_functions, custom_shape_calculators)
    topology.compile()
    onnx_model = convert_topology(topology, name, doc_string, target_opset, targeted_onnx)
    opsets = {d.domain: d.version for d in onnx_model.opset_import}
    if '' in opsets and opsets[''] < 9 and target_opset >= 9:
        # a bug?
        opsets[''] = 9
        del onnx_model.opset_import[:]
        for k, v in opsets.items():
            opset = onnx_model.opset_import.add()
            opset.domain = k
            opset.version = v
    return onnx_model



# SPDX-License-Identifier: Apache-2.0

import sys
import unittest
import numpy
import pandas
from pyspark import Row
from pyspark.ml.classification import NaiveBayes
from pyspark.ml.linalg import Vectors
from onnxmltools import convert_sparkml
from onnxmltools.convert.common.data_types import FloatTensorType
from onnx.defs import onnx_opset_version
from onnxconverter_common.onnx_ex import DEFAULT_OPSET_NUMBER
from tests.sparkml.sparkml_test_utils import save_data_models, run_onnx_model, compare_results
from tests.sparkml import SparkMlTestCase


TARGET_OPSET = min(DEFAULT_OPSET_NUMBER, onnx_opset_version())


class TestSparkmlNaiveBayes(SparkMlTestCase):

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_naive_bayes_bernoulli(self):
        data = self.spark.createDataFrame([
            Row(label=0.0, weight=0.1, features=Vectors.dense([0.0, 0.0])),
            Row(label=0.0, weight=0.5, features=Vectors.dense([0.0, 1.0])),
            Row(label=1.0, weight=1.0, features=Vectors.dense([1.0, 0.0]))])
        nb = NaiveBayes(smoothing=1.0, modelType="bernoulli", weightCol="weight")
        model = nb.fit(data)
        feature_count = data.select('features').first()[0].size
        model_onnx = convert_sparkml(model, 'Sparkml NaiveBayes Bernoulli',
                                     [('features', FloatTensorType([None, feature_count]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
            ]
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlNaiveBayesBernoulli")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)

    @unittest.skipIf(sys.version_info < (3, 8),
                     reason="pickle fails on python 3.7")
    def test_naive_bayes_multinomial(self):
        data = self.spark.createDataFrame([
            Row(label=0.0, weight=0.1, features=Vectors.dense([0.0, 0.0])),
            Row(label=0.0, weight=0.5, features=Vectors.dense([0.0, 1.0])),
            Row(label=1.0, weight=1.0, features=Vectors.dense([1.0, 0.0]))])
        nb = NaiveBayes(smoothing=1.0, modelType="multinomial", weightCol="weight")
        model = nb.fit(data)
        feature_count = data.select('features').first()[0].size
        model_onnx = convert_sparkml(model, 'Sparkml NaiveBayes Multinomial',
                                     [('features', FloatTensorType([None, feature_count]))],
                                     target_opset=TARGET_OPSET)
        self.assertTrue(model_onnx is not None)

        # run the model
        predicted = model.transform(data)
        expected = [
            predicted.toPandas().prediction.values.astype(numpy.float32),
            predicted.toPandas().probability.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
            ]
        data_np = data.toPandas().features.apply(lambda x: pandas.Series(x.toArray())).values.astype(numpy.float32)
        paths = save_data_models(data_np, expected, model, model_onnx, basename="SparkmlNaiveBayesMultinomial")
        onnx_model_path = paths[-1]
        output, output_shapes = run_onnx_model(['prediction', 'probability'], data_np, onnx_model_path)
        compare_results(expected, output, decimal=5)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: Apache-2.0

from ...common._registration import register_converter, register_shape_calculator
from ...common.data_types import StringTensorType, FloatTensorType
from ...common.utils import check_input_and_output_numbers, check_input_and_output_types
from ...common._topology import Operator, Scope, ModelComponentContainer
from pyspark.ml.feature import CountVectorizerModel


def convert_count_vectorizer(scope: Scope, operator: Operator, container: ModelComponentContainer):
    op: CountVectorizerModel = operator.raw_operator
    vocab, minTF, binary = op.vocabulary, op.getOrDefault("minTF"), op.getOrDefault("binary")

    if minTF < 1.0:
        raise NotImplementedError("Converting to ONNX for CountVectorizerModel is not supported when minTF < 1.0")

    min_opset = 9
    if not binary:
        # If binary is False, then we need the ThresholdedRelu operator which is only available since opset 10.
        min_opset = 10

    if container.target_opset < min_opset:
        raise NotImplementedError(
            f"Converting to ONNX for CountVectorizerModel is not supported in opset < {min_opset}"
        )

    # Create a TfIdfVectorizer node with gram length set to 1 and mode set to "TF".
    vectorizer_output_variable_name = scope.get_unique_variable_name("vectorizer_output")
    tfIdfVectorizer_attrs = {
        "name": scope.get_unique_operator_name("tfIdfVectorizer"),
        "min_gram_length": 1,
        "max_gram_length": 1,
        "max_skip_count": 0,
        "mode": "TF",
        "ngram_counts": [0],
        "ngram_indexes": [*range(len(vocab))],
        "pool_strings": vocab,
    }

    container.add_node(
        op_type="TfIdfVectorizer",
        inputs=[operator.inputs[0].full_name],
        outputs=[vectorizer_output_variable_name],
        op_version=9,
        **tfIdfVectorizer_attrs,
    )

    # In Spark's CountVectorizerModel, the comparison with minTF is inclusive,
    # but in ThresholdedRelu (or Binarizer) node, the comparison with `alpha` (or `threshold`) is exclusive.
    # So, we need to subtract epsilon from minTF to make the comparison with `alpha` (or `threshold`) effectively inclusive.
    epsilon = 1e-6
    if binary:
        # Create a Binarizer node with threshold set to minTF - epsilon.
        container.add_node(
            op_type="Binarizer",
            inputs=[vectorizer_output_variable_name],
            outputs=[operator.outputs[0].full_name],
            op_version=1,
            op_domain="ai.onnx.ml",
            threshold=minTF - epsilon,
        )
    else:
        # Create a ThresholdedRelu node with alpha set to minTF - epsilon
        container.add_node(
            op_type="ThresholdedRelu",
            inputs=[vectorizer_output_variable_name],
            outputs=[operator.outputs[0].full_name],
            op_version=10,
            alpha=minTF - epsilon,
        )


register_converter("pyspark.ml.feature.CountVectorizerModel", convert_count_vectorizer)


def calculate_count_vectorizer_output_shapes(operator):
    check_input_and_output_numbers(operator, output_count_range=1)
    check_input_and_output_types(operator, good_input_types=[StringTensorType])

    N = operator.inputs[0].type.shape[0]
    C = len(operator.raw_operator.vocabulary)
    operator.outputs[0].type = FloatTensorType([N, C])


register_shape_calculator("pyspark.ml.feature.CountVectorizerModel", calculate_count_vectorizer_output_shapes)
